import re
import sys
import os
from .functions import *
################################################################################################
############################ Static (Do NOT change this code) ##################################
################################################################################################

global global_dic
global_dic = {}
global functions_pool

#####################################################################################################
########### ADD THE IMPLEMENTATION OF YOUR FUNCTIONS HERE FOLLOWING THE EXAMPLES ####################
#####################################################################################################

functions_pool = {"toLowerCase":"","toUpperCase":"","toUpperCaseURL":"",
                "replaceValue":"","concat2":"","uuid":"","helloworld":"",
                "escape":"","schema":"","string_replace":"",
                "parseURL":""}


## Define your functions here following examples below, the column "names" from the csv files 
## that you aim to use as the input parameters of functions are only required to be provided 
## as the keys of "global_dic"
def toLowerCase(): 
    return str(global_dic["valueParam"]).lower()

def toUpperCase(): 
    return str(global_dic["valueParam"]).upper()

def helloworld(): 
    return "Hello World!"

def string_replace(): 
    return global_dic["valueParam"].replace(global_dic["param_find"],global_dic["param_replace"])

def parseURL():
    parsed = {}
    parsed["protocolOutput"] = global_dic["stringParameter"].split("://")[0]
    if "#" in global_dic["stringParameter"]:
        parsed["stringOutput"] = global_dic["stringParameter"].split("://")[1].split("#")[1]
        parsed["domainOutput"] = global_dic["stringParameter"].split("://")[1].split("#")[0]
    else:
        parsed["stringOutput"] = global_dic["stringParameter"].split("://")[1].split("/")[len(global_dic["stringParameter"].split("://")[1].split("/"))-1]
        replace_end = "/" + parsed["stringOutput"]
        parsed["domainOutput"] = global_dic["stringParameter"].split("://")[1].replace(replace_end,"")
    return parsed

def concat2():
    value1 = global_dic["value1"]
    value2 = global_dic["value2"]
    if bool(value1) and bool(value2):
        result = str(str(value1)+str(value2))
    else:
        result = ""  
    return(result)

def uuid():
    from uuid import uuid4
    return str(uuid4())

def escape():
    if global_dic["modeParam"] == 'html':
        import html
        return html.escape(global_dic["valueParam"])
    elif global_dic["modeParam"] == 'url':
        import urllib.parse
        return urllib.parse.quote(global_dic["valueParam"])
    else:
        raise ValueError("Invalid mode. Use 'html' for HTML escaping or 'url' for URL escaping.")

def toUpperCaseURL():
    url_lower = global_dic["str"].lower()

    if url_lower.startswith('https://'):
        return global_dic["str"].upper()
    elif url_lower.startswith('http://'):
        return global_dic["str"].upper()

    # else:
    return f'http://{encode_char(global_dic["str"].upper())}'

def schema(): 
    return "https://schema.org/" + encode_char(global_dic["stringParameter"])
################################################################################################
############################ Static (Do NOT change this code) ##################################
################################################################################################

def execute_function(row,header,dic):
    if "#" in dic["function"]:
        func = dic["function"].split("#")[1]
    else:
        func = dic["function"].split("/")[len(dic["function"].split("/"))-1]
    if func in functions_pool:
        global global_dic
        global_dic = execution_dic(row,header,dic)
        if global_dic == None:
            print("Error when executing function")
            return None
        else:
            return eval(func + "()")             
    else:
        print("Invalid function")
        print("Aborting...")
        sys.exit(1)

def execution_dic(row,header,dic):
    output = {}
    for inputs in dic["inputs"]:
        if isinstance(inputs,list):
            if "constant" not in inputs:
                if "reference" in inputs[1]:
                    if isinstance(row,dict):
                        output[inputs[2]] = row[inputs[0]]
                    else:
                        output[inputs[2]] = row[header.index(inputs[0])]
                elif "template" in inputs:
                    if isinstance(row,dict):
                        output[inputs[2]] = string_substitution(inputs[0], "{(.+?)}", row, "subject", "yes", "None")
                    else:
                        output[inputs[2]] = string_substitution_array(inputs[0], "{(.+?)}", row, header, "subject", "yes")
            else:
                output[inputs[2]] = inputs[0]
        else:
            if "#" in inputs:
                param = inputs.split("#")[1]
            else:
                param = inputs.split("/")[len(inputs.split("/"))-1]
            if "constant" != dic["inputs"][inputs]["type"]:
                if "reference" == dic["inputs"][inputs]["type"]:
                    if isinstance(row,dict):
                        if dic["inputs"][inputs]["value"] in row:
                            output[param] = row[dic["inputs"][inputs]["value"]]
                        else:
                            return None
                    else:
                        if dic["inputs"][inputs]["value"] in header:
                            output[param] = row[header.index(dic["inputs"][inputs]["value"])]
                        else:
                            return None
                elif "template" == dic["inputs"][inputs]["type"]:
                    if isinstance(row,dict):
                        output[param] = string_substitution(dic["inputs"][inputs]["value"], "{(.+?)}", row, "subject", "yes", "None")
                    else:
                        output[param] = string_substitution_array(dic["inputs"][inputs]["value"], "{(.+?)}", row, header, "subject", "yes")
                    if output[param] == None:
                        return None
            else:
               output[param] = dic["inputs"][inputs]["value"] 
    return output


import re
import sys
import os
from .fnml_functions import *

def new_inner_function(row,function,triples_map):
    functions = []
    keys = []
    for func_map in triples_map.func_map_list:
        if func_map.func_map_id == function:
            for param in func_map.parameters:
                if func_map.parameters[param]["type"] == "function":
                    for fm in triples_map.func_map_list:
                        if fm.func_map_id == func_map.parameters[param]["value"]:
                            functions.append(func_map.parameters[param]["value"])
                    func_map.parameters[param]["type"] = "reference"    
                elif func_map.parameters[param]["type"] == "template":
                    if "{" in func_map.parameters[param]["value"]:
                        attr_list = func_map.parameters[param]["value"].split("{")
                        for attr in attr_list:
                            if "}" in attr:
                                keys.append(attr.split("}")[0])
                elif func_map.parameters[param]["type"] == "reference":
                    keys.append(func_map.parameters[param]["value"])
            if functions:
                temp_row = {}
                for func in functions:
                    value = new_inner_function(row,func,triples_map)
                    temp_row[func] = value
                for key in keys:
                    temp_row[key] = row[key]
                current_func = {"inputs":func_map.parameters, 
                                "function":func_map.name}
                return execute_function(temp_row,None,current_func)
            else:
                current_func = {"inputs":func_map.parameters, 
                                "function":func_map.name}
                return execute_function(row,None,current_func)  


def inner_function(row,dic,triples_map_list):

    functions = []
    keys = []
    for attr in dic["inputs"]:
        if ("reference function" in attr[1]):
            functions.append(attr[0])
        elif "template" in attr[1]:
            for value in attr[0].split("{"):
                if "}" in value:
                   keys.append(value.split("}")[0]) 
        elif "constant" not in attr[1]:
            keys.append(attr[0])
    if functions:
        temp_dics = {}
        for function in functions:
            for tp in triples_map_list:
                if tp.triples_map_id == function:
                    temp_dic = create_dictionary(tp)
                    current_func = {"inputs":temp_dic["inputs"], 
                                    "function":temp_dic["executes"],
                                    "func_par":temp_dic,
                                    "termType":True}
                    temp_dics[function] = current_func
        temp_row = {}
        for dics in temp_dics:
            value = inner_function(row,temp_dics[dics],triples_map_list)
            temp_row[dics] = value
        for key in keys:
            temp_row[key] = row[key]
        return execute_function(temp_row,None,dic)
    else:
        return execute_function(row,None,dic)

def inner_values(row,dic,triples_map_list):
    values = ""
    for inputs in dic["inputs"]:
        if "reference" == inputs[1]:
            values += str(row[inputs[0]])
        elif "template" == inputs[1]:
            for string in inputs[0].split("{"):
                if "}" in string:
                    values += str(row[string.split("}")[0]])
        elif "reference function" == inputs[1]:
            temp_dics = {}
            for tp in triples_map_list:
                if tp.triples_map_id == inputs[0]:
                    temp_dic = create_dictionary(tp)
                    current_func = {"inputs":temp_dic["inputs"], 
                                    "function":temp_dic["executes"],
                                    "func_par":temp_dic,
                                    "termType":True}
                    values += inner_values(row,temp_dic,triples_map_list)
    return values

def inner_function_exists(inner_func, inner_functions):
    for inner_function in inner_functions:
        if inner_func["id"] in inner_function["id"]:
            return False
    return True

def create_dictionary(triple_map):
    dic = {}
    inputs = []
    for tp in triple_map.predicate_object_maps_list:
        if "#" in tp.predicate_map.value:
            key = tp.predicate_map.value.split("#")[1]
            tp_type = tp.predicate_map.mapping_type
        elif "/" in tp.predicate_map.value:
            key = tp.predicate_map.value.split("/")[len(tp.predicate_map.value.split("/"))-1]
            tp_type = tp.predicate_map.mapping_type
        if "constant" in tp.object_map.mapping_type:
            value = tp.object_map.value
            tp_type = tp.object_map.mapping_type
        if "template" in tp.object_map.mapping_type:
            value = tp.object_map.value
            tp_type = tp.object_map.mapping_type
        elif "executes" in tp.predicate_map.value:
            if "#" in tp.object_map.value:
                value = tp.object_map.value.split("#")[1]
                tp_type = tp.object_map.mapping_type
            elif "/" in tp.object_map.value:
                value = tp.object_map.value.split("/")[len(tp.object_map.value.split("/"))-1]
                tp_type = tp.object_map.mapping_type
        else:
            value = tp.object_map.value
            tp_type = tp.object_map.mapping_type

        dic.update({key : value})
        if (key != "executes") and ([value,tp_type,key] not in inputs):
            inputs.append([value,tp_type,key])

    dic["inputs"] = inputs
    return dic


import os
import re
import datetime
import sys
import urllib
import math
from .functions import *
global inner_join_table
inner_join_table = {}
global general_predicates
general_predicates = {"http://www.w3.org/2000/01/rdf-schema#subClassOf": "",
                      "http://www.w3.org/2002/07/owl#sameAs": "",
                      "http://www.w3.org/2000/01/rdf-schema#seeAlso": "",
                      "http://www.w3.org/2000/01/rdf-schema#subPropertyOf": ""}

def inner_hash_maker(parent_data, parent_subject, child_object, quoted, triples_map_list):
    global blank_message
    hash_table = {}
    for row in parent_data:
        if quoted == "":
            if child_object.parent[0] in row.keys():
                if row[child_object.parent[0]] in hash_table:
                    if duplicate == "yes":
                        if parent_subject.subject_map.subject_mapping_type == "reference":
                            value = string_substitution(parent_subject.subject_map.value, ".+", row, "object", ignore,
                                                        parent_subject.iterator)
                            if value != None:
                                if "http" in value and "<" not in value:
                                    value = "<" + value[1:-1] + ">"
                                elif "http" in value and "<" in value:
                                    value = value[1:-1]
                            if value not in hash_table[row[child_object.parent[0]]]:
                                hash_table[row[child_object.parent[0]]].update({value: "object"})
                        else:
                            if string_substitution(parent_subject.subject_map.value, "{(.+?)}", row, "object", ignore,
                                                   parent_subject.iterator) != None:
                                value = string_substitution(parent_subject.subject_map.value, "{(.+?)}", row, "object",
                                                            ignore, parent_subject.iterator)
                                if value != None:
                                    if parent_subject.subject_map.term_type != None:
                                        if "BlankNode" in parent_subject.subject_map.term_type:
                                            if "/" in value:
                                                value = "_:" + encode_char(value.replace("/", "2F")).replace("%", "")
                                                if "." in value:
                                                    value = value.replace(".", "2E")
                                                if blank_message:
                                                    logger.warning(
                                                        "Incorrect format for Blank Nodes. \"/\" will be replace with \"2F\".")
                                                    blank_message = False
                                            else:
                                                value = "_:" + encode_char(value).replace("%", "")
                                                if "." in value:
                                                    value = value.replace(".", "2E")
                                    else:
                                        value = "<" + value + ">"
                                    hash_table[row[child_object.parent[0]]].update({value: "object"})
                    else:
                        if parent_subject.subject_map.subject_mapping_type == "reference":
                            value = string_substitution(parent_subject.subject_map.value, ".+", row, "object", ignore,
                                                        parent_subject.iterator)
                            if "http" in value and "<" not in value:
                                value = "<" + value[1:-1] + ">"
                            elif "http" in value and "<" in value:
                                value = value[1:-1]
                            hash_table[row[child_object.parent[0]]].update({value: "object"})
                        else:
                            value = string_substitution(parent_subject.subject_map.value, "{(.+?)}", row, "object", ignore,
                                                        parent_subject.iterator)
                            if value != None:
                                if parent_subject.subject_map.term_type != None:
                                    if "BlankNode" in parent_subject.subject_map.term_type:
                                        if "/" in value:
                                            value = "_:" + encode_char(value.replace("/", "2F")).replace("%", "")
                                            if "." in value:
                                                value = value.replace(".", "2E")
                                            if blank_message:
                                                logger.warning(
                                                    "Incorrect format for Blank Nodes. \"/\" will be replace with \"2F\".")
                                                blank_message = False
                                        else:
                                            value = "_:" + encode_char(value).replace("%", "")
                                            if "." in value:
                                                value = value.replace(".", "2E")
                                else:
                                    value = "<" + value + ">"
                                hash_table[row[child_object.parent[0]]].update({value: "object"})

                else:
                    if parent_subject.subject_map.subject_mapping_type == "reference":
                        value = string_substitution(parent_subject.subject_map.value, ".+", row, "object", ignore,
                                                    parent_subject.iterator)
                        if value != None:
                            if "http" in value and "<" not in value:
                                value = "<" + value[1:-1] + ">"
                            elif "http" in value and "<" in value:
                                value = value[1:-1]
                        hash_table.update({row[child_object.parent[0]]: {value: "object"}})
                    else:
                        value = string_substitution(parent_subject.subject_map.value, "{(.+?)}", row, "object", ignore,
                                                    parent_subject.iterator)
                        if value != None:
                            if parent_subject.subject_map.term_type != None:
                                if "BlankNode" in parent_subject.subject_map.term_type:
                                    if "/" in value:
                                        value = "_:" + encode_char(value.replace("/", "2F")).replace("%", "")
                                        if "." in value:
                                            value = value.replace(".", "2E")
                                        if blank_message:
                                            logger.warning(
                                                "Incorrect format for Blank Nodes. \"/\" will be replace with \"2F\".")
                                            blank_message = False
                                    else:
                                        value = "_:" + encode_char(value).replace("%", "")
                                        if "." in value:
                                            value = value.replace(".", "2E")
                            else:
                                value = "<" + value + ">"
                            hash_table.update({row[child_object.parent[0]]: {value: "object"}})
        else:
            for triples in inner_semantify_file(parent_subject, triples_map_list, ",", row, base):
                if triples != None:
                    if isinstance(child_object.parent,list):
                        parent = child_object.parent[0]
                    else:
                        parent = child_object.parent
                    if row[parent] in hash_table:
                        if duplicate == "yes":
                            if triples not in hash_table[row[parent]]:
                                hash_table[row[parent]].update({triples : "subject"})
                        else:
                            hash_table[row[parent]].update({triples : "subject"})
                    else:
                        hash_table.update({row[parent] : {triples : "subject"}})
    if isinstance(child_object.child,list):
        join_table.update({parent_subject.triples_map_id + "_" + child_object.child[0] : hash_table})
    else:
        join_table.update({"quoted_" + parent_subject.triples_map_id + "_" + child_object.child[0] : hash_table})

def inner_semantify_file(triples_map, triples_map_list, delimiter, row, base):
	object_list = []
	subject_list = []
	triples_list = []

	if triples_map.subject_map.subject_mapping_type == "template":
		subject_value = string_substitution(triples_map.subject_map.value, "{(.+?)}", row, "subject", "yes", triples_map.iterator)
		if triples_map.subject_map.term_type is None:
			if triples_map.subject_map.condition == "":

				try:
					subject = "<" + subject_value + ">"
				except:
					subject = None

			else:
			#	field, condition = condition_separetor(triples_map.subject_map.condition)
			#	if row[field] == condition:
				try:
					subject = "<" + subject_value  + ">"
				except:
					subject = None 
		else:
			if "IRI" in triples_map.subject_map.term_type:
				subject_value = string_substitution(triples_map.subject_map.value, "{(.+?)}", row, "subject", "yes", triples_map.iterator)
				if triples_map.subject_map.condition == "":

					try:
						if "http" not in subject_value:
							subject = "<" + base + subject_value + ">"
						else:
							subject = "<" + encode_char(subject_value) + ">"
					except:
						subject = None

				else:
				#	field, condition = condition_separetor(triples_map.subject_map.condition)
				#	if row[field] == condition:
					try:
						if "http" not in subject_value:
							subject = subject = "<" + base + subject_value + ">"
						else:
							subject = "<" + subject_value + ">"
					except:
						subject = None 

			elif "BlankNode" in triples_map.subject_map.term_type:
				if triples_map.subject_map.condition == "":
					try:
						if "/" in subject_value:
							subject  = "_:" + encode_char(subject_value.replace("/","2F")).replace("%","")
							if "." in subject:
								subject = subject.replace(".","2E")
							if blank_message:
								print("Incorrect format for Blank Nodes. \"/\" will be replace with \"2F\".")
								blank_message = False
						else:
							subject = "_:" + encode_char(subject_value).replace("%","")
							if "." in subject:
								subject = subject.replace(".","2E")
					except:
						subject = None

				else:
				#	field, condition = condition_separetor(triples_map.subject_map.condition)
				#	if row[field] == condition:
					try:
						subject = "_:" + subject_value  
					except:
						subject = None
			elif "Literal" in triples_map.subject_map.term_type:
				subject = None			
			else:
				if triples_map.subject_map.condition == "":

					try:
						subject = "<" + subject_value + ">"
					except:
						subject = None

				else:
				#	field, condition = condition_separetor(triples_map.subject_map.condition)
				#	if row[field] == condition:
					try:
						subject = "<" + subject_value + ">"
					except:
						subject = None 
	elif "reference" in triples_map.subject_map.subject_mapping_type:
		subject_value = string_substitution(triples_map.subject_map.value, ".+", row, "subject","yes" , triples_map.iterator)
		if subject_value != None:
			subject_value = subject_value[1:-1]
			if triples_map.subject_map.condition == "":
				if " " not in subject_value:
					if "BlankNode" in triples_map.subject_map.term_type:
						subject = "_:" + subject_value
					else:
						if "http" not in subject_value:
							subject = "<" + base + subject_value + ">"
						else:
							subject = "<" + subject_value + ">"
				else:
					subject = None

		else:
		#	field, condition = condition_separetor(triples_map.subject_map.condition)
		#	if row[field] == condition:
			try:
				if "http" not in subject_value:
					subject = "<" + base + subject_value + ">"
				else:
					subject = "<" + subject_value + ">"
			except:
				subject = None

	elif "constant" in triples_map.subject_map.subject_mapping_type:
		subject = "<" + subject_value + ">"

	elif "quoted triples map" in triples_map.subject_map.subject_mapping_type:
		for triples_map_element in triples_map_list:
			if triples_map_element.triples_map_id == triples_map.subject_map.value:
				if triples_map_element.data_source != triples_map.data_source:
					if triples_map.subject_map.parent[0] != None:
						if ("quoted_" + triples_map_element.triples_map_id + "_" + triples_map.subject_map.child[0]) not in join_table:
							if str(triples_map_element.file_format).lower() == "csv" or triples_map_element.file_format == "JSONPath":
								with open(str(triples_map_element.data_source), "r") as input_file_descriptor:
									if str(triples_map_element.file_format).lower() == "csv":
										data = csv.DictReader(input_file_descriptor, delimiter=',')
										hash_maker(data, triples_map_element, triples_map.subject_map, "quoted", triples_map_list)
									else:
										pass
						if row[triples_map.subject_map.child[0]] in join_table["quoted_" + triples_map_element.triples_map_id + "_" + triples_map.subject_map.child[0]]:
						    subject_list = join_table["quoted_" + triples_map_element.triples_map_id + "_" + triples_map.subject_map.child[0]][row[triples_map.subject_map.child[0]]]
				else:
					subject_list = inner_semantify_file(triples_map_element, triples_map_list, delimiter, row, base)
				subject = None

	else:
		if triples_map.subject_map.condition == "":

			try:
				subject = "\"" + triples_map.subject_map.value + "\""
			except:
				subject = None

		else:
		#	field, condition = condition_separetor(triples_map.subject_map.condition)
		#	if row[field] == condition:
			try:
				subject = "\"" + triples_map.subject_map.value + "\""
			except:
				subject = None


	if triples_map.subject_map.rdf_class != None and subject != None:
		predicate = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>"
		for rdf_class in triples_map.subject_map.rdf_class:
			if rdf_class != None:
				obj = "<{}>".format(rdf_class)
				rdf_type = subject + " " + predicate + " " + obj 
				for graph in triples_map.subject_map.graph:
					if graph != None and "defaultGraph" not in graph:
						if "{" in graph:	
							rdf_type = rdf_type[:-2] + " <" + string_substitution(graph, "{(.+?)}", row, "subject","yes", triples_map.iterator) + "> .\n"
						else:
							rdf_type = rdf_type[:-2] + " <" + graph + "> .\n"
				triples_list.append(rdf_type)

	
	for predicate_object_map in triples_map.predicate_object_maps_list:
		if predicate_object_map.predicate_map.mapping_type == "constant" or predicate_object_map.predicate_map.mapping_type == "constant shortcut":
			predicate = "<" + predicate_object_map.predicate_map.value + ">"
		elif predicate_object_map.predicate_map.mapping_type == "template":
			if predicate_object_map.predicate_map.condition != "":
					#field, condition = condition_separetor(predicate_object_map.predicate_map.condition)
					#if row[field] == condition:
					try:
						predicate = "<" + string_substitution(predicate_object_map.predicate_map.value, "{(.+?)}", row, "predicate","yes", triples_map.iterator) + ">"
					except:
						predicate = None
					#else:
					#	predicate = None
			else:
				try:
					predicate = "<" + string_substitution(predicate_object_map.predicate_map.value, "{(.+?)}", row, "predicate","yes", triples_map.iterator) + ">"
				except:
					predicate = None
		elif predicate_object_map.predicate_map.mapping_type == "reference":
			if predicate_object_map.predicate_map.condition != "":
				#field, condition = condition_separetor(predicate_object_map.predicate_map.condition)
				#if row[field] == condition:
				predicate = string_substitution(predicate_object_map.predicate_map.value, ".+", row, "predicate","yes", triples_map.iterator)
				#else:
				#	predicate = None
			else:
				predicate = string_substitution(predicate_object_map.predicate_map.value, ".+", row, "predicate","yes", triples_map.iterator)
			predicate = "<" + predicate[1:-1] + ">"
		else:
			predicate = None

		if predicate_object_map.object_map.mapping_type == "constant" or predicate_object_map.object_map.mapping_type == "constant shortcut":
			if "/" in predicate_object_map.object_map.value:
				object = "<" + predicate_object_map.object_map.value + ">"
			else:
				object = "\"" + predicate_object_map.object_map.value + "\""
			if predicate_object_map.object_map.datatype != None:
				object = "\"" + object[1:-1] + "\"" + "^^<{}>".format(predicate_object_map.object_map.datatype)
		elif predicate_object_map.object_map.mapping_type == "template":
			try:
				if predicate_object_map.object_map.term is None:
					object = "<" + string_substitution(predicate_object_map.object_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator) + ">"
				elif "IRI" in predicate_object_map.object_map.term:
					object = "<" + string_substitution(predicate_object_map.object_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator) + ">"
				elif "BlankNode" in predicate_object_map.object_map.term:
					object = "_:" + string_substitution(predicate_object_map.object_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator)
					if "/" in object:
						object  = object.replace("/","2F")
						if blank_message:
							print("Incorrect format for Blank Nodes. \"/\" will be replace with \"2F\".")
							blank_message = False
					if "." in object:
						object = object.replace(".","2E")
					object = encode_char(object)
				else:
					object = "\"" + string_substitution(predicate_object_map.object_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator) + "\""
					if predicate_object_map.object_map.datatype != None:
						object = "\"" + object[1:-1] + "\"" + "^^<{}>".format(predicate_object_map.object_map.datatype)
					elif predicate_object_map.object_map.language != None:
						if "spanish" in predicate_object_map.object_map.language or "es" in predicate_object_map.object_map.language :
							object += "@es"
						elif "english" in predicate_object_map.object_map.language or "en" in predicate_object_map.object_map.language :
							object += "@en"
						elif len(predicate_object_map.object_map.language) == 2:
							object += "@"+predicate_object_map.object_map.language
					elif predicate_object_map.object_map.language_map != None:
						lang = string_substitution(predicate_object_map.object_map.language_map, ".+", row, "object","yes", triples_map.iterator)
						if lang != None:
							object += "@" + string_substitution(predicate_object_map.object_map.language_map, ".+", row, "object","yes", triples_map.iterator)[1:-1] 
			except TypeError:
				object = None
		elif predicate_object_map.object_map.mapping_type == "reference":
			object = string_substitution(predicate_object_map.object_map.value, ".+", row, "object","yes", triples_map.iterator)
			if object != None:
				if "\\" in object[1:-1]:
					object = "\"" + object[1:-1].replace("\\","\\\\") + "\""
				if "'" in object[1:-1]:
					object = "\"" + object[1:-1].replace("'","\\\\'") + "\""
				if "\n" in object:
					object = object.replace("\n","\\n")
				if predicate_object_map.object_map.datatype != None:
					object = "\"" + object[1:-1] + "\"" + "^^<{}>".format(predicate_object_map.object_map.datatype)
				elif predicate_object_map.object_map.language != None:
					if "spanish" in predicate_object_map.object_map.language or "es" in predicate_object_map.object_map.language :
						object += "@es"
					elif "english" in predicate_object_map.object_map.language or "en" in predicate_object_map.object_map.language :
						object += "@en"
					elif len(predicate_object_map.object_map.language) == 2:
						object += "@"+predicate_object_map.object_map.language
				elif predicate_object_map.object_map.language_map != None:
					lang = string_substitution(predicate_object_map.object_map.language_map, ".+", row, "object","yes", triples_map.iterator)
					if lang != None:
						object += "@"+ string_substitution(predicate_object_map.object_map.language_map, ".+", row, "object","yes", triples_map.iterator)[1:-1]
				elif predicate_object_map.object_map.term != None:
					if "IRI" in predicate_object_map.object_map.term:
						if " " not in object:
							object = "\"" + object[1:-1].replace("\\\\'","'") + "\""
							object = "<" + encode_char(object[1:-1]) + ">"
						else:
							object = None
					elif "BlankNode" in predicate_object_map.object_map.term:
						if " " not in object:
							object = "_:" + object[1:-1]
						else:
							object = None 
		elif predicate_object_map.object_map.mapping_type == "parent triples map":
			if subject != None:
				for triples_map_inner in triples_map_list:
					if triples_map_inner.triples_map_id == predicate_object_map.object_map.value:
						if triples_map_inner.data_source != triples_map.data_source:
							if len(predicate_object_map.object_map.child) == 1:
								if (triples_map_inner.triples_map_id + "_" + predicate_object_map.object_map.child[0]) not in join_table:
									if str(triples_map_inner.file_format).lower() == "csv" or triples_map.file_format == "JSONPath":
										with open(str(triples_map_inner.data_source), "r") as input_file_descriptor:
											if str(triples_map.file_format).lower() == "csv":
												reader = pd.read_csv(str(triples_map_inner.data_source), dtype = str)#, encoding = "ISO-8859-1")
												reader = reader.where(pd.notnull(reader), None)
												reader = reader.drop_duplicates(keep ='first')
												data = reader.to_dict(orient='records')
												hash_maker(data, triples_map_inner, predicate_object_map.object_map)
											else:
												data = json.load(input_file_descriptor)
												if triples_map_inner.iterator:
													if triples_map_inner.iterator != "None" and triples_map_inner.iterator != "$.[*]":
														join_iterator(data, triples_map_inner.iterator, triples_map_inner, predicate_object_map.object_map)
													else:
														if isinstance(data, list):
															hash_maker(data, triples_map_inner, predicate_object_map.object_map)
														elif len(data) < 2:
															hash_maker(data[list(data.keys())[0]], triples_map_inner, predicate_object_map.object_map)
												else:
													if isinstance(data, list):
														hash_maker(data, triples_map_inner, predicate_object_map.object_map)
													elif len(data) < 2:
														hash_maker(data[list(data.keys())[0]], triples_map_inner, predicate_object_map.object_map)

									elif triples_map.file_format == "XPath":
										with open(str(triples_map.data_source), "r") as input_file_descriptor:
											child_tree = ET.parse(input_file_descriptor)
											child_root = child_tree.getroot()
											hash_maker_xml(child_root, triples_map_inner, predicate_object_map.object_map)								
									else:
										database, query_list = translate_sql(triples_map)
										db = connector.connect(host=host, port=int(port), user=user, password=password)
										cursor = db.cursor(buffered=True)
										cursor.execute("use " + database)
										for query in query_list:
											cursor.execute(query)
										hash_maker_array(cursor, triples_map_inner, predicate_object_map.object_map)

								if sublist(predicate_object_map.object_map.child,row.keys()):
									if child_list_value(predicate_object_map.object_map.child,row) in join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)]:
										object_list = join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)][child_list_value(predicate_object_map.object_map.child,row)]
									else:
										if no_update:
											if str(triples_map_inner.file_format).lower() == "csv" or triples_map_inner.file_format == "JSONPath":
												with open(str(triples_map_inner.data_source), "r") as input_file_descriptor:
													if str(triples_map_inner.file_format).lower() == "csv":
														reader = pd.read_csv(str(triples_map_inner.data_source), dtype = str)#, encoding = "ISO-8859-1")
														reader = reader.where(pd.notnull(reader), None)
														reader = reader.drop_duplicates(keep ='first')
														data = reader.to_dict(orient='records')
														hash_update(data, triples_map_inner, predicate_object_map.object_map, triples_map_inner.triples_map_id + "_" + predicate_object_map.object_map.child[0])
													else:
														data = json.load(input_file_descriptor)
														if triples_map_inner.iterator:
															if triples_map_inner.iterator != "None" and triples_map_inner.iterator != "$.[*]":
																join_iterator(data, triples_map_inner.iterator, triples_map_inner, predicate_object_map.object_map)
															else:
																if isinstance(data, list):
																	hash_maker(data, triples_map_inner, predicate_object_map.object_map)
																elif len(data) < 2:
																	hash_maker(data[list(data.keys())[0]], triples_map_inner, predicate_object_map.object_map)
														else:
															if isinstance(data, list):
																hash_maker(data, triples_map_inner, predicate_object_map.object_map)
															elif len(data) < 2:
																hash_maker(data[list(data.keys())[0]], triples_map_inner, predicate_object_map.object_map)
											if child_list_value(predicate_object_map.object_map.child,row) in join_table[triples_map_inner.triples_map_id + "_" + predicate_object_map.object_map.child[0]]:
												object_list = join_table[triples_map_inner.triples_map_id + "_" + predicate_object_map.object_map.child[0]][row[predicate_object_map.object_map.child[0]]]
											else:
												object_list = []
											no_update = False
								object = None
							else:
								if (triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)) not in join_table:
									if str(triples_map_inner.file_format).lower() == "csv" or triples_map_inner.file_format == "JSONPath":
										with open(str(triples_map_inner.data_source), "r") as input_file_descriptor:
											if str(triples_map_inner.file_format).lower() == "csv":
												data = csv.DictReader(input_file_descriptor, delimiter=delimiter)
												hash_maker_list(data, triples_map_inner, predicate_object_map.object_map)
											else:
												data = json.load(input_file_descriptor)
												if isinstance(data, list):
													hash_maker_list(data, triples_map_inner, predicate_object_map.object_map)
												elif len(data) < 2:
													hash_maker_list(data[list(data.keys())[0]], triples_map_inner, predicate_object_map.object_map)

									elif triples_map_inner.file_format == "XPath":
										with open(str(triples_map_inner.data_source), "r") as input_file_descriptor:
											child_tree = ET.parse(input_file_descriptor)
											child_root = child_tree.getroot()
											hash_maker_xml(child_root, triples_map_inner, predicate_object_map.object_map)						
									else:
										database, query_list = translate_sql(triples_map)
										db = connector.connect(host=host, port=int(port), user=user, password=password)
										cursor = db.cursor(buffered=True)
										cursor.execute("use " + database)
										for query in query_list:
											cursor.execute(query)
										hash_maker_array(cursor, triples_map_inner, predicate_object_map.object_map)
								if sublist(predicate_object_map.object_map.child,row.keys()):
									if child_list_value(predicate_object_map.object_map.child,row) in join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)]:
										object_list = join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)][child_list_value(predicate_object_map.object_map.child,row)]
									else:
										object_list = []
								object = None
						else:
							if predicate_object_map.object_map.parent != None:
								if predicate_object_map.object_map.parent[0] != predicate_object_map.object_map.child[0]:
									if (triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)) not in join_table:
										with open(str(triples_map_inner.data_source), "r") as input_file_descriptor:
											if str(triples_map_inner.file_format).lower() == "csv":
												parent_data = csv.DictReader(input_file_descriptor, delimiter=delimiter)
												hash_maker_list(parent_data, triples_map_inner, predicate_object_map.object_map)
											else:
												parent_data = json.load(input_file_descriptor)
												if isinstance(parent_data, list):
													hash_maker_list(parent_data, triples_map_inner, predicate_object_map.object_map)
												else:
													hash_maker_list(parent_data[list(parent_data.keys())[0]], triples_map_inner, predicate_object_map.object_map)
									if sublist(predicate_object_map.object_map.child,row.keys()):
										if child_list_value(predicate_object_map.object_map.child,row) in join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)]:
											object_list = join_table[triples_map_inner.triples_map_id + "_" + child_list(predicate_object_map.object_map.child)][child_list_value(predicate_object_map.object_map.child,row)]
										else:
											object_list = []
									object = None
								else:
									try:
										object = "<" + string_substitution(triples_map_inner.subject_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator) + ">"
									except TypeError:
										object = None
							else:
								try:
									object = "<" + string_substitution(triples_map_inner.subject_map.value, "{(.+?)}", row, "object","yes", triples_map.iterator) + ">"
								except TypeError:
									object = None
						break
					else:
						continue
			else:
				object = None
		elif "quoted triples map" in predicate_object_map.object_map.mapping_type:
			for triples_map_element in triples_map_list:
				if triples_map_element.triples_map_id == predicate_object_map.object_map.value:
					if triples_map_element.data_source != triples_map.data_source:
						if predicate_object_map.object_map.parent[0] != None:
							if ("quoted_" + triples_map_element.triples_map_id + "_" + predicate_object_map.object_map.child[0]) not in join_table:
								if str(triples_map_element.file_format).lower() == "csv" or triples_map_element.file_format == "JSONPath":
									with open(str(triples_map_element.data_source), "r") as input_file_descriptor:
										if str(triples_map_element.file_format).lower() == "csv":
											data = csv.DictReader(input_file_descriptor, delimiter=',')
											inner_hash_maker(data, triples_map_element, predicate_object_map.object_map, "quoted", triples_map_list)
										else:
											pass
							if row[predicate_object_map.object_map.child[0]] in join_table["quoted_" + triples_map_element.triples_map_id + "_" + predicate_object_map.object_map.child[0]]:
								object_list = join_table["quoted_" + triples_map_element.triples_map_id + "_" + predicate_object_map.object_map.child[0]][row[predicate_object_map.object_map.child[0]]]
					else:
						object_list = inner_semantify_file(triples_map_element, triples_map_list, delimiter, row, base)
					object = None
		else:
			object = None
		if predicate != None and object != None and subject != None:
			for graph in triples_map.subject_map.graph:
				triple = subject + " " + predicate + " " + object 
				if graph != None and "defaultGraph" not in graph:
					if "{" in graph:
						triple = triple[:-2] + " <" + string_substitution(graph, "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
					else:
						triple = triple[:-2] + " <" + graph + ">"
				triples_list.append(triple)
			if predicate[1:-1] in predicate_object_map.graph:
				triple = subject + " " + predicate + " " + object 
				if predicate_object_map.graph[predicate[1:-1]] != None and "defaultGraph" not in predicate_object_map.graph[predicate[1:-1]]:
					if "{" in predicate_object_map.graph[predicate[1:-1]]:
						triple = triple[:-2] + " <" + string_substitution(predicate_object_map.graph[predicate[1:-1]], "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
					else:
						triple = triple[:-2] + " <" + predicate_object_map.graph[predicate[1:-1]] + ">"
					triples_list.append(triple)
		elif predicate != None and subject != None and object_list:
			for obj in object_list:
				if obj != None:
					for graph in triples_map.subject_map.graph:
						if predicate_object_map.object_map.term != None:
							if "IRI" in predicate_object_map.object_map.term:
								triple = subject + " " + predicate + " <" + obj[1:-1] + ">"
							else:
								triple = subject + " " + predicate + " " + obj 
						else:
							if "quoted triples map" in predicate_object_map.object_map.mapping_type:
								triple = subject + " " + predicate + " <<" + obj + ">>"
							else:
								triple = subject + " " + predicate + " " + obj 
						if graph != None and "defaultGraph" not in graph:
							if "{" in graph:
								triple = triple[:-2] + " <" + string_substitution(graph, "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
							else:
								triple = triple[:-2] + " <" + graph + ">"
						triples_list.append(triple)

					if predicate[1:-1] in predicate_object_map.graph:
						if predicate_object_map.object_map.term != None:
							if "IRI" in predicate_object_map.object_map.term:
								triple = subject + " " + predicate + " <" + obj[1:-1] + ">"
							else:
								triple = subject + " " + predicate + " " + obj 
						else:
							triple = subject + " " + predicate + " " + obj 
						if predicate_object_map.graph[predicate[1:-1]] != None and "defaultGraph" not in predicate_object_map.graph[predicate[1:-1]]:
							if "{" in predicate_object_map.graph[predicate[1:-1]]:
								triple = triple[:-2] + " <" + string_substitution(predicate_object_map.graph[predicate[1:-1]], "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
							else:
								triple = triple[:-2] + " <" + predicate_object_map.graph[predicate[1:-1]] + ">"
							triples_list.append(triple)
			object_list = []
		elif predicate != None and object != None and subject_list:
			for subj in subject_list:
				if subj != None:
					for graph in triples_map.subject_map.graph:
						triple = "<< " + subj + ">> " + predicate + " " + object 
						if graph != None and "defaultGraph" not in graph:
							if "{" in graph:
								triple = triple[:-2] + " <" + string_substitution(graph, "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
							else:
								triple = triple[:-2] + " <" + graph + ">"
						triples_list.append(triple)

					if predicate[1:-1] in predicate_object_map.graph:
						triple = "<< " + subj + " >> " + predicate + " " + object 
						if predicate_object_map.graph[predicate[1:-1]] != None and "defaultGraph" not in predicate_object_map.graph[predicate[1:-1]]:
							if "{" in predicate_object_map.graph[predicate[1:-1]]:
								triple = triple[:-2] + " <" + string_substitution(predicate_object_map.graph[predicate[1:-1]], "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
							else:
								triple = triple[:-2] + " <" + predicate_object_map.graph[predicate[1:-1]] + ">"
							triples_list.append(triple)
			subject_list = []
		elif predicate != None and object_list and subject_list:
			for subj in subject_list:
				for obj in object_list:
					if subj != None:
						for graph in triples_map.subject_map.graph:
							if "quoted triples map" in predicate_object_map.object_map.mapping_type:
								triple = "<< " + subj + ">> " + predicate + " <<" + obj + ">>" 
							else:
								triple = "<< " + subj + ">> " + predicate + " " + obj 
							if graph != None and "defaultGraph" not in graph:
								if "{" in graph:
									triple = triple[:-2] + " <" + string_substitution(graph, "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
								else:
									triple = triple[:-2] + " <" + graph + ">"
							triples_list.append(triple)

						if predicate[1:-1] in predicate_object_map.graph:
							triple = "<< " + subj + ">> " + predicate + " " + obj 
							if predicate_object_map.graph[predicate[1:-1]] != None and "defaultGraph" not in predicate_object_map.graph[predicate[1:-1]]:
								if "{" in predicate_object_map.graph[predicate[1:-1]]:
									triple = triple[:-2] + " <" + string_substitution(predicate_object_map.graph[predicate[1:-1]], "{(.+?)}", row, "subject","yes", triples_map.iterator) + ">"
								else:
									triple = triple[:-2] + " <" + predicate_object_map.graph[predicate[1:-1]] + ">"
			subject_list = []
			object_list = []
		else:
			continue
	return triples_list